import LinearCode
import BicycleCode
import FiniteProjectiveGeometryCode
import FiniteEuclideanGeometryCode

LinearCode=LinearCode.LinearCode
BicycleCode=BicycleCode.BicycleCode
FiniteProjectiveGeometryCode=FiniteProjectiveGeometryCode.FiniteProjectiveGeometryCode
FiniteEuclideanGeometryCode=FiniteEuclideanGeometryCode.FiniteEuclideanGeometryCode
